# SleepFade (c) 2007 by Peter Watkins (peterw@tux.org) 
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, which should be included with this software.

package Plugins::SleepFade::Plugin;

use strict;

use Slim::Control::Command;
use Slim::Control::Request;
use Slim::Utils::DateTime;
use Slim::Utils::Strings qw (string);
use Slim::Utils::Misc qw( msg );
use POSIX qw(strftime);


# ------------------------- SETTINGS -------------------------------
my $minimumFadeTime = 900;	# if sleep set to less than this, will not fade (900 = 15 minutes; 
				# sleep duration must ALSO be a multiple of 900, so "end of song" works for long recordings)
my $sleepEarly = 60;		# how much "early" to hit volume 0 (to avoid Slimserver's last-minute fade-out)
my $defaultFadeMode = 0;	# 0 == do not fade; 1 = linear; 3 == more at start; 0.333 == more at end; 0.125 == mostly at end
my $defaultVolumeRestore = 1;	# 0 == do not restore volume; 1 = do restore volume
my $maxWait = 5;		# check for volume changes at least every $maxWait seconds
# ------------------------- SETTINGS -------------------------------

# globals
my $originalSleepCommand; 	# need to keep track of original function
my $fadeConfigName = 'plugin-SleepFade-fade';
my $restoreVolumeConfigName = 'plugin-SleepFade-restore-volume';
my $restoreVolumeLevelName = 'plugin-SleepFade-restore-volume-level';

my %menuItems;		# up/down choice arrays
$menuItems{''} =  [ 'PLUGIN_SLEEPFADE_MODE', 'PLUGIN_SLEEPFADE_RESTORE_VOLUME'];
$menuItems{'0'} =  [ 'PLUGIN_SLEEPFADE_MODE_OFF', 'PLUGIN_SLEEPFADE_MODE_LINEAR', 'PLUGIN_SLEEPFADE_MODE_STEEP', 'PLUGIN_SLEEPFADE_MODE_FLAT', 'PLUGIN_SLEEPFADE_MODE_VERY_FLAT' ]; 
$menuItems{'1'} =  [ 'PLUGIN_SLEEPFADE_RESTORE_VOLUME_OFF', 'PLUGIN_SLEEPFADE_RESTORE_VOLUME_ON' ]; 
			
my %modeValues = (
	'PLUGIN_SLEEPFADE_RESTORE_VOLUME_OFF' => '0', 
	'PLUGIN_SLEEPFADE_RESTORE_VOLUME_ON' => '1', 
	'PLUGIN_SLEEPFADE_MODE_OFF' => '0', 
	'PLUGIN_SLEEPFADE_MODE_LINEAR' => '1', 
	'PLUGIN_SLEEPFADE_MODE_STEEP' => '3', 
	'PLUGIN_SLEEPFADE_MODE_FLAT' => '0.333',
	'PLUGIN_SLEEPFADE_MODE_VERY_FLAT' => '0.125',
);

my @settingNames = ($fadeConfigName, $restoreVolumeConfigName);

# per-player
# for fade operation:
my %startTime;		# when sleep started
my %sleepDuration;	# how much time set for this sleep (constant)
my %volumeIncrement;	# Transporters have 100 steps; older players have 40
my %failures;		# watch out for volume-locked players
# for menu operation
my %xPos;		# menu left/right pos
my %yPos;		# menu up/down pos
my %menuTrail;		# set of x/y positions

$Plugins::SleepFade::Plugin::apiVersion = 1.0;

use vars qw($VERSION);
$VERSION = &vlVersion();

# our i18n/l10n strings
sub strings() { 
  	return '
PLUGIN_SLEEPFADE_RESTORE_VOLUME
	EN	Volume level restoration

PLUGIN_SLEEPFADE_RESTORE_VOLUME_OFF
	EN	Remain silent after sleeping

PLUGIN_SLEEPFADE_RESTORE_VOLUME_ON
	EN	Restore volume after sleeping

PLUGIN_SLEEPFADE_MODE
	EN	Sleep volume fading mode

PLUGIN_SLEEPFADE_MODE_OFF
	EN	Do not change the volume level

PLUGIN_SLEEPFADE_MODE_LINEAR
	EN	Linear volume decrease

PLUGIN_SLEEPFADE_MODE_STEEP
	EN	Fade volume more initially

PLUGIN_SLEEPFADE_MODE_FLAT
	EN	Fade volume more towards end

PLUGIN_SLEEPFADE_MODE_VERY_FLAT
	EN	Fade volume mostly at end

PLUGIN_SLEEPFADE
	EN	SleepFade
'};

# whether we're enabled (the callback remains in the function 
# stack after "disabling" the plugin, so we need to keep track of this)
my $pluginEnabled = 0;

# whether we've already subscribed to events
my $callbackset = 0;

# setup routine
sub initPlugin {
	# note that we should act
	$pluginEnabled = 1;
	if ( $callbackset == 0 ) {
		$::d_plugins && &Plugin_msg("registering callback\n");
		Slim::Control::Request::subscribe( \&newPlayerReset, [['client']],[['new']]);
                $::d_plugins && &Plugin_msg("wrapping button functions\n");
                $originalSleepCommand = Slim::Control::Request::addDispatch(['sleep','_newvalue'],[1, 0, 0, \&SleepFade_sleepCommand]);
                if ( (!defined($originalSleepCommand)) || (ref($originalSleepCommand ) ne 'CODE') ) {
                        $::d_plugins && &Plugin_msg("problem wrapping IR command!\n");
                }
		$callbackset = 1;
	}
	$::d_plugins && &Plugin_msg("enabled\n");
}

sub shutdownPlugin {
	# we should not act
	$pluginEnabled = 0;
	$::d_plugins && &Plugin_msg("disabled\n");
}

sub enabled {
	if ( substr($::VERSION,0,3) gt '6.5' ) {
		&Plugin_msg("This plugin has not been tested with SlimServer newer than 6.5.x\n");
	}
	return ($::VERSION ge '6.5');
}

sub vlVersion() {
	my $rcsVersion = '$Revision: 1.2 $';
	$rcsVersion =~ s/.*:\s*([0-9\.]*).*$/$1/;
	return $rcsVersion;
}

sub getDisplayName() {
	return 'PLUGIN_SLEEPFADE';
}

# our IR wrapper function
sub SleepFade_sleepCommand {
        my @args = @_;
        my $request = $args[0];
        my $newvalue = $request->getParam('_newvalue');
	my $client = $request->client();
	# call the real command to change the value
        my $rc = &$originalSleepCommand(@args);
	if ( defined($client) && ($pluginEnabled == 1) ) {
		my $id = $client->id();
#$::d_plugins && &Plugin_msg("$id, new value \"$newvalue\"\n");
		# init player config
		&initPlayerConfig($client,$id);
		Slim::Utils::Timers::killTimers($client, \&decreaseVolume);
		Slim::Utils::Timers::killTimers($client, \&clientReset);
		my $restorePref = Slim::Utils::Prefs::clientGet($client, $restoreVolumeConfigName);
		my $originalVolume = Slim::Utils::Prefs::clientGet($client, $restoreVolumeLevelName);
		if ( ($restorePref == 1) && ($newvalue == 0) && ($originalVolume ne '') ) {
			# restore volume
			#Slim::Player::Client::volume($client,$originalVolume,undef);
			Slim::Control::Request::executeRequest($client, ['mixer', 'volume', $originalVolume], undef, undef);
			Slim::Utils::Prefs::clientSet($client, $restoreVolumeLevelName,'');
		}
		my $fadePref = Slim::Utils::Prefs::clientGet($client, $fadeConfigName);
		# if player fades when sleeping
		if ( $fadePref > 0  ) {
			# if sleep not disabled or too short, and if a multiple of 15 minutes
			if ( ($newvalue >= $minimumFadeTime) && ( ($newvalue % 900) == 0) ) {
				# set current volume
				my $volNow = $client->volume();
				$volNow = $volumeIncrement{$id} * int( $volNow / $volumeIncrement{$id});
				#Slim::Player::Client::volume($client,$volNow,undef);
				Slim::Control::Request::executeRequest($client, ['mixer', 'volume', $volNow], undef, undef);
				Slim::Utils::Prefs::clientSet($client, $restoreVolumeLevelName,$volNow);
				# set sleepDuration{$id}
				$sleepDuration{$id} = $newvalue;
				# Slimserver messes with the volume in the last 60 seconds; let's stop that
				if ( $sleepEarly > 0 ) {
					$::d_plugins && &Plugin_msg("$id: dropping to volume zero $sleepEarly seconds early to better control volume fade\n");
					$sleepDuration{$id} -= $sleepEarly;
				}
				# set start time
				$startTime{$id} = Time::HiRes::time();
				# set timer to reset volume once asleep
				Slim::Utils::Timers::setTimer($client, (Time::HiRes::time() + $newvalue + 1), \&clientReset);
				# set timer to decrease volume
				&setDecreaseTimer($client,$id);
			} elsif ( $newvalue > 0 ) {
				$::d_plugins && &Plugin_msg("$id: sleeping in $newvalue seconds, but only fade for timers of $minimumFadeTime or more\n");
			} elsif ( $newvalue == 0 ) {
				$::d_plugins && &Plugin_msg("$id: sleep cancelled, will not restore volume\n");
				Slim::Utils::Prefs::clientSet($client, $restoreVolumeLevelName, '');
			}
		} else {
			if ($originalVolume ne '') {
				$::d_plugins && &Plugin_msg("$id: not fading, will not restore volume\n");
				Slim::Utils::Prefs::clientSet($client, $restoreVolumeLevelName, '');
			}
		}
	}
        return $rc;
}


sub setDecreaseTimer($$) {
	my ($client, $id) = @_;
	if (!$client->power()) {
		# powered off
		&clientReset($client);
		return;
	}
	my $fadePref = Slim::Utils::Prefs::clientGet($client, $fadeConfigName);
	if ( $fadePref == 0 ) { 
		# no longer fading
		$::d_plugins && &Plugin_msg("$id no longer going to sleep\n");
		# don't reset the volume
		Slim::Utils::Prefs::clientSet($client, $restoreVolumeLevelName,'');
		return; 
	}
	my $volNow = $client->volume();
	# round to nearest increment
	$volNow = $volumeIncrement{$id} * int( $volNow / $volumeIncrement{$id});
	my $nextVol = &getVolumeForTime($client, $id, Time::HiRes::time()) - $volumeIncrement{$id};
	$nextVol = ( $nextVol < 0 ) ? 0 : $nextVol;
	my $changeTime = &getTimeForVolume($client,$id,undef);
	if (! defined($changeTime) ) {
		# nothing to do!
		return;
	}
	my $delta = $changeTime - Time::HiRes::time();
	# don't run too soon
	if ( $delta < 0.5 ) {
$::d_plugins && &Plugin_msg("$id, decrease volume from $volNow to $nextVol in $delta seconds is too soon\n");
		$changeTime = Time::HiRes::time() + 0.5;
	} elsif ( $delta > $maxWait ) {
		$changeTime = Time::HiRes::time() + $maxWait;
	}
	if ( $volNow < 0 ) {
		# we should never get here!
		# no rush!
		$changeTime += 5;
	}
	$delta = int(100 * ($changeTime - Time::HiRes::time())) / 100;
#$::d_plugins && &Plugin_msg("$id, decrease volume from $volNow to $nextVol in $delta seconds\n");
	Slim::Utils::Timers::setTimer($client, $changeTime, \&decreaseVolume);
}

sub getVolumeForTime {
	my ($client,$id,$now) = @_;
	# divide by zero protection
	if ( $sleepDuration{$id} == 0 ) {
		$::d_plugins && &Plugin_msg("$id, division by zero prevented (duration)\n");
		return $client->volume();
	}
	my $fadePref = Slim::Utils::Prefs::clientGet($client, $fadeConfigName);
	my $originalVolume = Slim::Utils::Prefs::clientGet($client, $restoreVolumeLevelName);
	my $volNext = $originalVolume * ( (1 - (($now - $startTime{$id}) / $sleepDuration{$id})) ** $fadePref );
#$::d_plugins && &Plugin_msg("$id, volume should be $volNext\n");
	#$volNext = $volumeIncrement{$id} * int( ($volNext / $volumeIncrement{$id}) + 0.25 );
	# NaN fix in case asked about time later than "0 volume" time
	# (alternate might be "if ( $now > &getTimeForVolume($client,$id,0) ) { $volNext = 0; }") but this works
	if ( $volNext != $volNext ) { $volNext = 0; }
	return $volNext;
}

sub getTimeForVolume {
	my ($client,$id,$volume) = @_;
	if (! defined($volume) ) {
		$volume = &getVolumeForTime($client,$id,Time::HiRes::time());
		$volume = ( $volume < 0 ) ? 0 : $volume;
		$volume = $volumeIncrement{$id} * int( $volume / $volumeIncrement{$id} );
		if ( abs($client->volume() - $volume) < $volumeIncrement{$id} ) {
			# find next step down
			$volume -= $volumeIncrement{$id};
		}
	}
	if ( $volume < 0 ) {
		return undef;
	}
	my $originalVolume = Slim::Utils::Prefs::clientGet($client, $restoreVolumeLevelName);
	# divide by zero protection
	if ( $originalVolume == 0 ) {
		$::d_plugins && &Plugin_msg("$id, division by zero prevented (original volume)\n");
		return undef;
	}
	my $fadePref = Slim::Utils::Prefs::clientGet($client, $fadeConfigName);
	my $when = $sleepDuration{$id} + $startTime{$id} - ( $sleepDuration{$id} * ( ($volume / $originalVolume) ** (1 / $fadePref)));
	return $when;
}

sub getVolumeIncrement($) {
	my $client = shift;
	# BUG (?) use increment of 2.5 for older players?
	#if ( $client->isa( "Slim::Player::SLIMP3") ) { return 2.5; }
	return 1.0;
}

sub decreaseVolume {
	my $client = shift;
	my $id = $client->id();
	if ( ($client->sleepTime() == undef) || ($client->sleepTime() < Time::HiRes::time()) ) {
		# no longer sleeping; bail out
		$::d_plugins && &Plugin_msg("$id no longer sleeping\n");
		&clientReset($client);
		return;
	}
	my $volNow = $client->volume();
	# round to nearest increment
	$volNow = $volumeIncrement{$id} * int( $volNow / $volumeIncrement{$id});
	my $nextVol = &getVolumeForTime($client,$id,Time::HiRes::time());
	$nextVol = ( $nextVol < 0 ) ? 0 : $nextVol;
	my $volRemainder = $nextVol - int($nextVol);
	if ( 
		($nextVol < $volNow) &&
		(
			( ($volNow - $nextVol) >= $volumeIncrement{$id} )
		)
	 ) {
		$nextVol = $volumeIncrement{$id} * int( ($nextVol / $volumeIncrement{$id}) + 0.999);
$::d_plugins && &Plugin_msg("$id, decreasing volume from $volNow to $nextVol\n");
		#Slim::Player::Client::volume($client,$nextVol,undef);
		Slim::Control::Request::executeRequest($client, ['mixer', 'volume', $nextVol], undef, undef);
		if ( $client->volume() > $nextVol ) {
			# can't change volume -- VolumeLock???
			++$failures{$id};
			if ( $failures{$id} > 5 ) {
				# give up
				$::d_plugins && &Plugin_msg("$id: unable to change volume from from $volNow to $nextVol; leaving volume intact\n");
				return;
			}
		} else {
			$failures{$id} = 0;
		}
	}
	if ( $client->volume() < 1 ) {
		# no sense continuing
		$::d_plugins && &Plugin_msg("$id: powering off player\n");
		# now remove power
		Slim::Control::Request::executeRequest($client, ['power', '_newvalue', 0], undef, undef);
		# reset the volume now that the player is powered down
		&clientReset($client);
		# cancel sleep?
		Slim::Control::Request::executeRequest($client, ['sleep', '_newvalue', 0], undef, undef);
		return;
	}
	# set timer for next change
	&setDecreaseTimer($client, $id);
}

sub initPlayerConfig($$) {
	my ($client,$id) = @_;
	my $fadePref = Slim::Utils::Prefs::clientGet($client, $fadeConfigName);
	if ( (!defined($fadePref)) || ($fadePref eq '') ) {
		$::d_plugins && &Plugin_msg("$id: set $fadeConfigName to $defaultFadeMode\n");
		Slim::Utils::Prefs::clientSet($client, $fadeConfigName, $defaultFadeMode);
	}
	my $restorePref = Slim::Utils::Prefs::clientGet($client, $restoreVolumeConfigName);
	if ( (!defined($restorePref)) || ($restorePref eq '') ) {
		$::d_plugins && &Plugin_msg("$id: set $restoreVolumeConfigName to $defaultVolumeRestore\n");
		Slim::Utils::Prefs::clientSet($client, $restoreVolumeConfigName, $defaultVolumeRestore);
	}
	my $originalVolume = Slim::Utils::Prefs::clientGet($client, $restoreVolumeLevelName);
	if (! defined($originalVolume) ) {
		Slim::Utils::Prefs::clientSet($client, $restoreVolumeLevelName, '');
	}
	# player increment
	$volumeIncrement{$id} = &getVolumeIncrement($client);
#$::d_plugins && &Plugin_msg("$id has steps of $volumeIncrement{$id}\n");
}

sub newPlayerReset {
	my $request = shift;
	$::d_plugins && &Plugin_msg("newPlayerReset running\n");
	my $client = $request->client();
	if ( defined($client) && ($pluginEnabled == 1) ) {
		&clientReset($client);
	}
}

sub clientReset {
	my $client = shift;
	my $id = $client->id();
	if ( defined($id) ) {
		# reset volume if needed
		if ( ($client->sleepTime() == undef) || ($client->sleepTime() < Time::HiRes::time()) ) {
			my $restorePref = Slim::Utils::Prefs::clientGet($client, $restoreVolumeConfigName);
			my $originalVolume = Slim::Utils::Prefs::clientGet($client, $restoreVolumeLevelName);
			if ( ($restorePref == 1) && ($originalVolume ne '') ) {
				# restore volume
				$::d_plugins && &Plugin_msg("$id: restoring volume to $originalVolume\n");
				#Slim::Player::Client::volume($client,$originalVolume,undef);
				Slim::Control::Request::executeRequest($client, ['mixer', 'volume', $originalVolume], undef, undef);
				Slim::Utils::Prefs::clientSet($client, $restoreVolumeLevelName, '');
			}
		}
	}
}

sub Plugin_msg {
	my $message = shift;
	msg("SleepFade: $message");
}

# player UI code

sub addMenu {
	return "PLUGINS";
}

sub setMode {
        my $client = shift;
	$client->lines(\&displayLines);
}

our %menuFunctions = (
        # normal up/right/left/down navigation:
        'left' => sub  {
                my $client = shift;
		my $button = shift;
		if ( $button ne 'left' ) { return; }
		my $id = $client->id();
		&initPlayerConfig($client,$id);
		if ( &moveXRight($id,-1,$client) < 0 ) {
                	Slim::Buttons::Common::popModeRight($client);
		} else {
			$client->lines(\&displayLines);
			$client->update();
		}
        },
        'right' => sub  {
                my $client = shift;
		my $button = shift;
		if ( $button ne 'right' ) { return; }
		my $id = $client->id();
		&initPlayerConfig($client,$id);
		if ( &moveXRight($id,1,$client) < 0 ) {
                	$client->bumpRight($client);
		} else {
			$client->lines(\&displayLines);
			$client->update();
		}
        },
	# 'up' also handles 'up_repeat'
        'up' => sub  {
                my $client = shift;
		my $button = shift;
		if ( $button ne 'up' ) { return; }
		my $id = $client->id();
		&initPlayerConfig($client,$id);
		&moveYDown($id,-1,$client);
		$client->lines(\&displayLines);
		$client->update();
        },
	# 'down' also handles 'down_repeat'
        'down' => sub  {
                my $client = shift;
		my $button = shift;
		if ( $button ne 'down' ) { return; }
		my $id = $client->id();
		&initPlayerConfig($client,$id);
		&moveYDown($id,1,$client);
		$client->lines(\&displayLines);
		$client->update();
        },
);

sub moveXRight {
	my ($id,$amount,$client) = @_;
	&initMenuPos($id);
	$xPos{$id} += $amount;
	if ( $xPos{$id} < 0 ) { 
		$xPos{$id} = 0; 
		return -1;
	}
	if ( $xPos{$id} > 1) { 
		$xPos{$id} = 1; 
		return -1;
	}
	if ( $xPos{$id} == 1) { 
		# what's our value?
		my $confName = $settingNames[$yPos{$id}];
		my $confValue = Slim::Utils::Prefs::clientGet($client, $confName);
		my @choiceArray = @{$menuItems{substr($menuTrail{$id},0,1)}};
		# default to zero
		$yPos{$id} = 0;
		for (my $i = 0; $i < scalar(@choiceArray); ++$i) {
			my $thisVal = $modeValues{$choiceArray[$i]};
			if ( $thisVal eq $confValue ) { $yPos{$id} = $i; }
		}
	} else {
		if ( (length($menuTrail{$id}) -1) > $xPos{$id} ) {
 			$yPos{$id} = substr($menuTrail{$id},$xPos{$id},1) + 0;
		} else {
 			$yPos{$id} = 0;
		}
	}
	if ( $amount > 0 ) {
		# added
		$menuTrail{$id} .= $yPos{$id};
	} else {
		#removed
		$menuTrail{$id} =~ s/(\d)$//;
	}
	return $xPos{$id};
}

sub setThisYPos {
	my $id = shift;
	my $pos = $yPos{$id};
	$menuTrail{$id} =~ s/^(.*)(\d)$/$1$pos/;
}

sub moveYDown {
	my ($id,$amount,$client) = @_;
	&initMenuPos($id);
	my $maxPos;
	my $update = 0;
	if ( $xPos{$id} > 0 ) { $update = 1; }
	my @yArray = &getYArray($id);
	$maxPos = scalar(@yArray) - 1;
	$yPos{$id} += $amount;
	if ( $yPos{$id} < 0 ) { $yPos{$id} = $maxPos; }
	if ( defined($maxPos) && ($yPos{$id} > $maxPos) ) { 
		$yPos{$id} = 0; 
	}
	&setThisYPos($id);
	# update prefs?
	if ( $update == 1 ) {
		#&updatePref($id,$client);
		# what is first yPos?
		my $settingPos = substr($menuTrail{$id},0,1);
		my $confName = $settingNames[$settingPos];
		my @choiceArray = @{$menuItems{substr($menuTrail{$id},0,1)}};
		my $confValString = $choiceArray[$yPos{$id}];
		my $confValue = $modeValues{$confValString};
#$::d_plugins && &Plugin_msg("$id: set $confName to $confValue\n");
		Slim::Utils::Prefs::clientSet($client, $confName, $confValue);
	}
}

sub getYArray {
	my $id = shift;
	my $key = $menuTrail{$id};
	$key =~ s/.$//;
	my @yArray = @{$menuItems{$key}};
	return @yArray;
}

sub displayLines {
	my $client = shift;
	my $id = $client->id();
	&initMenuPos($id);
	my ($x, $y) = ($xPos{$id}, $yPos{$id});
	my $line1 = string('PLUGIN_SLEEPFADE');
	my @yArray = &getYArray($id);
	my $line2 = string($yArray[$yPos{$id}]);
        return($line1,$line2);
}

sub initMenuPos {
	my $id = shift;
	if (! defined($xPos{$id}) ) {
		$xPos{$id} = 0;
		$yPos{$id} = 0;
		$menuTrail{$id} = "0";
	}
}

sub getFunctions {
        return \%menuFunctions;
}

1;
